/*******************************************************************************
    Copyright (c) 2015 NVidia Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
*******************************************************************************/


#ifndef _NVLINK_COMMON_H_
#define _NVLINK_COMMON_H_

#ifdef __cplusplus
extern "C" {
#endif

#include "nvtypes.h"
#include "nvCpuUuid.h"

#ifndef NULL
#define NULL ((void *)0)
#endif

#define NVL_DBG_INFO       0x0
#define NVL_DBG_SETUP      0x1
#define NVL_DBG_USERERRORS 0x2
#define NVL_DBG_WARNINGS   0x3
#define NVL_DBG_ERRORS     0x4

#define NVL_FREE(x) nvl_free((void *)x)

typedef int                 NvlStatus;

// nvlink pci bar information
struct nvlink_pci_bar_info
{
    NvU64  busAddress;
    NvU64  baseAddr;
    NvU64  barSize;
    NvU32  offset;
    void  *pBar;
};

#define MAX_NVLINK_BARS    2

// nvlink pci information
struct nvlink_pci_info
{
    NvU16 domain;
    NvU8  bus;
    NvU8  device;
    NvU8  function;
    NvU16 pciDeviceId;

    // PCI Bar Information
    struct nvlink_pci_bar_info bars[MAX_NVLINK_BARS];
};

struct nvlink_conn_info
{
    NvU16  domain;
    NvU16  bus;
    NvU16  device;
    NvU16  function;
    NvU32  pciDeviceId;
    NvU8   devUuid[NV_UUID_LEN];
    NvU64  deviceType;
    NvU32  linkNumber;
    NvBool bConnected;
};


// Typedefs
typedef struct nvlink_pci_bar_info nvlink_pci_bar_info;
typedef struct nvlink_pci_info nvlink_pci_info;
typedef struct nvlink_conn_info nvlink_conn_info;

void        nvl_print(int, const char *, ...);
void       *nvl_malloc(NvLength);
void        nvl_free(void *);
char       *nvl_strcpy(char *, const char *);
NvLength    nvl_strlen(const char *);
int         nvl_strcmp(const char *, const char *);
int         nvl_snprintf(char *, NvLength, const char *, ...);
void       *nvl_memset(void *, int, NvLength);
void       *nvl_memcpy(void *, void *, NvLength);
int         nvlMemRd32(const volatile void *);
void        nvlMemWr32(volatile void *, unsigned int);
int         nvlMemRd64(const volatile void *);
void        nvlMemWr64(volatile void *, unsigned long long);
NvU8        nvlPciCfgRd08(NvU16, NvU8, NvU8, NvU8, NvU32); 
NvU16       nvlPciCfgRd16(NvU16, NvU8, NvU8, NvU8, NvU32);
NvU32       nvlPciCfgRd32(NvU16, NvU8, NvU8, NvU8, NvU32);
void        nvlPciCfgWr08(NvU16, NvU8, NvU8, NvU8, NvU32, NvU8);
void        nvlPciCfgWr16(NvU16, NvU8, NvU8, NvU8, NvU32, NvU16);
void        nvlPciCfgWr32(NvU16, NvU8, NvU8, NvU8, NvU32, NvU32);
void        nvl_assert(int expression);
void        nvl_sleep(unsigned int ms);

/*
 * Initializes core lib and does all that is needed
 * to access NVLINK functionality on the current platform.
 */
NvlStatus nvlink_lib_initialize(void);

/*
 * Frees any related resources and then unloads core lib. 
 */
NvlStatus nvlink_lib_unload(void);

/*
 * @Brief: Loads PCI devices matching the Ebridge profile into the driver.
 *
 * @Description :
 *
 * @returns                 NVL_SUCCESS if the action succeeded
 *                          An NVL error code otherwise.
 *                          Output prints will show if registration failed
 *                          for some devices.
 */
NvlStatus ebridge_lib_find_devices(void);

NvlStatus ebridge_lib_initialize_device_pci_bus(nvlink_pci_info *info);

NvlStatus ebridge_lib_initialize_device_bar_info(nvlink_pci_info *info);

NvlStatus ebridge_lib_shutdown_device_bar_info(nvlink_pci_info *info);



/*
 * @Brief: Loads PCI devices matching the NPU profile into the driver.
 *
 * @Description :
 *
 * @returns                 NVL_SUCCESS if the action succeeded
 *                          An NVL error code otherwise.
 *                          Output prints will show if registration failed
 *                          for some devices.
 */
NvlStatus ibmnpu_lib_find_devices(void);

NvlStatus ibmnpu_lib_initialize_device_pci_bus(nvlink_pci_info *info);

NvlStatus ibmnpu_lib_initialize_device_bar_info(nvlink_pci_info *info);

NvlStatus ibmnpu_lib_shutdown_device_bar_info(nvlink_pci_info *info);

#ifdef __cplusplus
}
#endif

#endif //_NVLINK_COMMON_H_
